
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#if defined(__WIN32__) || defined(_WIN32) || defined(_WIN64) || defined(__WINDOWS__)
#include <windows.h>
#include <tchar.h>
#endif

#include "QstCfg.h"
#include "QstCmd.h"
#include "QstComm.h"

#include "AccessQst.h"
#include "UsageStr.h"

#define MAX_FAN_CTRL_FIELD 79

/****************************************************************************/
/* Module Variables                                                         */
/****************************************************************************/

static QST_GENERIC_CMD                      stGenericCmd;
static QST_GET_SUBSYSTEM_STATUS_RSP         stStatusRsp;
static BOOL                                 bConfig = FALSE;
static BOOL                                 bQstCommInit = FALSE;
double fMinTemp=45;
double fMaxTemp=67;

/****************************************************************************/
/* DisplayError() - Displays an error message. Has support for the embedded */
/* QST error codes that are generated by the AccessQst module.              */
/****************************************************************************/

typedef enum
{
   QST_ERROR,
   WINDOWS_ERROR,
   ERRNO_ERROR,
   PICK_ERROR,

} E_ERROR_TYPE;

static const char * const pszQstError[] =
{
   "QST Successful",
   "QST Rejected Command: Command Unsupported",
   "QST Rejected Command: Capability Locked",
   "QST Rejected Command: Invalid Parameter",
   "QST Rejected Command: Invalid Version",
   "QST Command Execution Failed: Communication Error",
   "QST Command Execution Failed: Sensor Error",
   "QST Command Execution Failed: No Memory",
   "QST Command Execution Failed: No Resources",
   "QST Rejected Command: Invalid In This Context",
   "QST Rejected Command: Improper Command Size",
   "QST Rejected Command: Improper Response Size",
   "Unknown Error Code"
};

static void DisplayError( char *pszMessage, E_ERROR_TYPE eError, DWORD dwError )
{
   printf( "\n*** %s!!\n", pszMessage );

   switch( eError )
   {
   case QST_ERROR:

      if( IS_QST_ERROR( dwError ) )
      {
         if( dwError == ERROR_QST_NOT_CONFIGURED )
         {
            puts( "   QST Error = 0x0100 (Not Configured)" );
            break;
         }

         dwError -= ERROR_QST_BASE;
      }

      if( dwError <= QST_CMD_REJECTED_RSP_SIZE )
         printf( "   QST Error = 0x%04X (%s)\n\n", dwError, pszQstError[dwError] );
      else if( dwError )
         printf( "   QST Error = 0x%04X (Unknown Error)\n\n", dwError );

      break;

   case ERRNO_ERROR:

      if( errno )
         printf( "   ERRNO = %d (%s)\n\n", errno, strerror(errno) );

      break;

   case PICK_ERROR:

#if defined(__MSDOS__) || defined(__LINUX__) || defined(__SOLARIS__)

      if( errno )
      {
         if( IS_QST_ERROR( errno ) )
         {
            if( errno <= ERROR_QST_IMPROPER_RESPONSE_SIZE )
               printf( "   QST Error = 0x%04X (%s)\n", errno - ERROR_QST_BASE, pszQstError[errno - ERROR_QST_BASE] );
            else if( errno == ERROR_QST_NOT_CONFIGURED )
               puts(   "   QST Error = 0x0100 (Not Configured)" );
            else
               printf( "   QST Error = 0x%04X (Unknown Error)\n", errno - ERROR_QST_BASE );
         }
         else
            printf( "   ERRNO = %d (%s)\n\n", errno, strerror(errno) );
      }

      break;

#elif defined(__WIN32__)

      dwError = GetLastError();

   case WINDOWS_ERROR:

      if( dwError )
      {
         if( IS_QST_ERROR( dwError ) )
         {
            if( dwError <= ERROR_QST_IMPROPER_RESPONSE_SIZE )
               printf( "   QST Error = 0x%04X (%s)\n", dwError - ERROR_QST_BASE, pszQstError[dwError - ERROR_QST_BASE] );
            else if( dwError == ERROR_QST_NOT_CONFIGURED )
               printf( "   QST Error = 0x%04X (Not Configured)\n", dwError - ERROR_QST_BASE );
            else
               printf( "   QST Error = 0x%04X (Unknown Error)\n", dwError - ERROR_QST_BASE );
         }
         else
         {
            char   *pszMsgBuf;
            size_t tLen;

            FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                           NULL, dwError, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR)&pszMsgBuf, 0, NULL );

            tLen = strlen(pszMsgBuf);

            if( tLen > 2 )
            {
               if( pszMsgBuf[tLen - 2] == '\r' )
                  pszMsgBuf[tLen - 2] = '\0';

               else if( pszMsgBuf[tLen - 1] == '\n' )
                  pszMsgBuf[tLen - 1] = '\0';

               printf( "   Windows Error = 0x%08X (%s)\n", dwError, pszMsgBuf );
            }
            else
               printf( "   Windows Error = 0x%08X (Unknown Error)\n", dwError );

            LocalFree( pszMsgBuf );
         }
      }

      break;

#endif

   }
}

#define uint16_t unsigned short

static BOOL SetFan( uint16_t speed )
{
   int     i,iIndex, iUsage, iStatus, iFans = GetFanCountQst();
   float   fValue;
   THRESH  stThresh;

   for( iIndex = 0; iIndex < iFans; iIndex++ )
   {
      if( (iStatus = GetFanHealthQst( iIndex )) != -1 )
      {
         if( (iUsage = GetFanUsageQst( iIndex )) != -1 )
         {
            if( (fValue = GetFanReadingQst( iIndex )) != -1 )
            {
               if( GetFanThreshQst( iIndex, &stThresh ) )
               {
                    if(iUsage==1)
                         SetDutyManualQst(iIndex, speed);
		    continue;
               }
            }
         }
      }

      DisplayError( "Unable to obtain Fan Speed Sensor information", PICK_ERROR, 0 );
      return( FALSE );
   }

   return( TRUE );
}

static BOOL TestTempSensors( void )
{
   int     iIndex, iUsage, iStatus, iTemps = GetTempCountQst();
   float   fValue;
   THRESH  stThresh;

   for( iIndex = 0; iIndex < iTemps; iIndex++ )
   {
      if( (iStatus = GetTempHealthQst( iIndex )) != -1 )
      {
         if( (iUsage = GetTempUsageQst( iIndex )) != -1 )
         {
            if( (fValue  = GetTempReadingQst( iIndex )) != -1 )
            {
               if( GetTempThreshQst( iIndex, &stThresh ) )
               {
		  if((iUsage==1)||(iUsage==2)) {
                      uint16_t speed;
                      fValue+=100;
		      if(fValue<=fMinTemp) {
                         speed=0;
                      } else
		      if(fValue>=fMaxTemp) {
                         speed=100;
                      } else {
		         speed=(uint16_t)((double)(fValue-fMinTemp)/(double)(fMaxTemp-fMinTemp)*100);
                      } 
                      SetFan(speed);
                  }
                  continue;
               }
            }
         }
      }

      DisplayError( "Unable to obtain Temperature Sensor information", PICK_ERROR, 0 );
      return( FALSE );
   }

   return( TRUE );
}

/* Daemonize. Source: http://www-theorie.physik.unizh.ch/~dpotter/howto/daemonize (public domain) */
static void daemonize(void) {
	pid_t pid, sid;

	/* already a daemon */
	if (getppid() == 1)
		return;

	/* Fork off the parent process */
	pid = fork();
	if (pid < 0) {
		exit(EXIT_FAILURE);
	}
	/* If we got a good PID, then we can exit the parent process. */
	if (pid > 0) {
		exit(EXIT_SUCCESS);
	}

	/* At this point we are executing as the child process */

	/* Change the file mode mask */
	umask(0);

	/* Create a new SID for the child process */
	sid = setsid();
	if (sid < 0) {
		exit(EXIT_FAILURE);
	}

	/* Change the current working directory.  This prevents the current
	 directory from being locked; hence not being able to remove it. */
	if ((chdir("/")) < 0) {
		exit(EXIT_FAILURE);
	}

	/* Redirect standard files to /dev/null */
	void* r;
	r = freopen("/dev/null", "r", stdin);
	r = freopen("/dev/null", "w", stdout);
	r = freopen("/dev/null", "w", stderr);
}

/****************************************************************************/
/* main() - Mainline for program                                            */
/****************************************************************************/

int main( int iArgs, char *pszArg[] )
{
   puts( "\nIntel(R) Quiet System Technology Fun Control" );
   puts( "Copyleft (C) 2010, Vanya Batalin.\n" );

   if( !InitializeQst() )
   {

#if defined(__WIN32__)
      DWORD dwError = GetLastError();
#else
      UINT32 dwError = (UINT32)errno;
#endif

      if( dwError == ERROR_QST_NOT_CONFIGURED )
      {
         puts( "The QST Subsystem is NOT configured!" );

#if defined(DYNAMIC_DLL_LOADING) || defined(__MSDOS__)

         // Initialize basic communications

         if( !QstInitialize() )
         {
            DisplayError( "Unable to communicate with QST Subsystem", PICK_ERROR, 0 );

#if defined(__WIN32__)
            dwError = GetLastError();
#else
            dwError = (UINT32)errno;
#endif

            return( (int)dwError );
         }

         bQstCommInit = TRUE;

#endif

      }
      else
      {
         DisplayError( "Unable to communicate with QST Subsystem", PICK_ERROR, 0 );
         return( (int)dwError );
      }
   }

   daemonize();
   while(TestTempSensors()) {
      sleep(1);
   }

#if defined(DYNAMIC_DLL_LOADING) || defined(__MSDOS__)

   if( bQstCommInit )
      QstCleanup();
   else

#endif

      CleanupQst();

   return( 0 );

#if _MSC_VER > 1000
   UNREFERENCED_PARAMETER( iArgs );
   UNREFERENCED_PARAMETER( pszArg );
#endif

}

