/*
 * Part of Intel(R) Manageability Engine Interface Linux driver
 *
 * Copyright (c) 2003 - 2008 Intel Corp.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    substantially similar to the "NO WARRANTY" disclaimer below
 *    ("Disclaimer") and any redistribution must be conditioned upon
 *    including a substantially similar Disclaimer requirement for further
 *    binary redistribution.
 * 3. Neither the names of the above-listed copyright holders nor the names
 *    of any contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * NO WARRANTY
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 */

#include "heci.h"


const __u8 heci_start_wd_params[] = { 0x02, 0x12, 0x13, 0x10 };
const __u8 heci_stop_wd_params[] = { 0x02, 0x02, 0x14, 0x10 };

const __u8 heci_wd_state_independence_msg[3][4] = {
	{0x05, 0x02, 0x51, 0x10},
	{0x05, 0x02, 0x52, 0x10},
	{0x07, 0x02, 0x01, 0x10}
};

const struct guid heci_asf_guid = {
	0x75B30CD6, 0xA29E, 0x4AF7,
	{0xA7, 0x12, 0xE6, 0x17, 0x43, 0x93, 0xC8, 0xA6}
};
const struct guid heci_wd_guid = {
	0x05B79A6F, 0x4628, 0x4D7F,
	{0x89, 0x9D, 0xA9, 0x15, 0x14, 0xCB, 0x32, 0xAB}
};
const struct guid heci_pthi_guid = {
	0x12f80028, 0xb4b7, 0x4b2d,
	{0xac, 0xa8, 0x46, 0xe0, 0xff, 0x65, 0x81, 0x4c}
};


/*
 *  heci init function prototypes
 */
static void heci_check_asf_mode(struct iamt_heci_device *dev);
static int host_start_message(struct iamt_heci_device *dev);
static int host_enum_clients_message(struct iamt_heci_device *dev);
static int allocate_me_clients_storage(struct iamt_heci_device *dev);
static int host_init_wd(struct iamt_heci_device *dev);
static int host_init_iamthif(struct iamt_heci_device *dev);
static int heci_wait_msg_timeout(struct iamt_heci_device *dev,
				       long timeout);


/**
 * heci_initialize_list - Sets up a queue list.
 *
 * @list: An instance of our list structure
 */
void heci_initialize_list(struct heci_io_list *list)
{
	/* initialize our queue list */
	INIT_LIST_HEAD(&list->cb_list);
}

/**
 * heci_flush_list - remove list entry belong to heci connection.
 *
 * @list:  An instance of our list structure
 * @conn: connection data of the file object
 * @priv_cb: callback block
 *
 * returns 1 if flushed, 0 if not
 */
int heci_flush_list(struct heci_io_list *list,
		struct heci_connection *conn,
		struct heci_cb_private *priv_cb)
{
	struct heci_cb_private *priv_cb_pos = NULL;
	struct heci_cb_private *priv_cb_next = NULL;
	int ret = 0;

	if ((list == NULL) || (conn == NULL))
		return 0;

	if (list_empty(&list->cb_list))
		return 0;

	list_for_each_entry_safe(priv_cb_pos, priv_cb_next,
					&list->cb_list, cb_list) {
		if (priv_cb_pos->conn == conn) {
			if ((priv_cb == NULL) || (priv_cb_pos == priv_cb)) {
				list_del_init(&priv_cb_pos->cb_list);
				ret = 1;
			}
		}
	}
	return ret;
}

/**
 * heci_reset_file_priv - resets private file structure.
 *
 * @file_ext: private data of the file object to be reset
 * @file: the file structure
 */
static void heci_reset_file_priv(struct heci_file_priv *file_ext,
				   struct file *file)
{
	INIT_LIST_HEAD(&file_ext->link);
	init_waitqueue_head(&file_ext->rx_wait);
	init_waitqueue_head(&file_ext->tx_wait);
	DBG("file_ext->rx_wait =%p\n", &file_ext->rx_wait);
	atomic_set(&file_ext->sm_state, 0);
	file_ext->file = file;
	atomic_set(&file_ext->r_state, HECI_IDLE);
	atomic_set(&file_ext->w_state, HECI_IDLE);
}

/**
 * heci_reset_heci_connection - resets connection structure.
 *
 * @conn: connection data of the file object to be reset
 * @file: the file structure
 */
static void heci_reset_heci_connection(struct heci_connection *conn,
				   struct file *file)
{
	INIT_LIST_HEAD(&conn->link);
	init_waitqueue_head(&conn->cc_wait);
	atomic_set(&conn->state, HECI_FILE_INITIALIZING);
	atomic_set(&conn->refcnt, 0);
	conn->reading_state = HECI_IDLE;
	conn->wr_state = HECI_IDLE;
	conn->file = file;
}

/**
 * heci_init_file_priv - initializes private file structure.
 *
 * @file_ext: private data of the file object to be initialized
 * @file: the file structure
 */
static void heci_init_file_priv(struct heci_file_priv *file_ext,
				   struct file *file)
{
	memset(file_ext, 0, sizeof(struct heci_file_priv));
	heci_reset_file_priv(file_ext, file);
}

/**
 * heci_init_heci_connection - initializes connection structure.
 *
 * @conn: connection data of the file object to be initialized
 * @file: the file structure
 */
static void heci_init_heci_connection(struct heci_connection *conn,
				   struct file *file)
{
	memset(conn, 0, sizeof(struct heci_connection));
	spin_lock_init(&conn->file_lock);
	heci_reset_heci_connection(conn, file);
}

/**
 * heci_alloc_file_priv - allocates a private file structure and set it up.
 * @file: the file structure
 *
 * returns  The allocated file connection data or NULL on failure
 */
struct heci_file_priv *heci_alloc_file_priv(struct file *file)
{
	struct heci_file_priv *file_ext;

	file_ext = kmalloc(sizeof(struct heci_file_priv), GFP_KERNEL);
	if (file_ext == NULL)
		return NULL;

	heci_init_file_priv(file_ext, file);

	return file_ext;
}

/**
 * heci_alloc_heci_connection - allocates connection structure and set it up.
 * @file: the file structure
 *
 * returns  The allocated file connection data or NULL on failure
 */
struct heci_connection *heci_alloc_heci_connection(struct file *file)
{
	struct heci_connection *conn;

	conn = kmalloc(sizeof(struct heci_connection), GFP_KERNEL);
	if (conn == NULL)
		return NULL;

	heci_init_heci_connection(conn, file);

	return conn;
}

/**
 * heci_put_heci_connection - free connection structure.
 *
 * @conn: connection data of the file object
 */
void heci_put_heci_connection(struct heci_connection *conn)
{
	if (atomic_dec_and_test(&conn->refcnt))
		kfree(conn);
}

/**
 * init_heci_device - allocates and initializes the heci device structure
 *
 * @pdev: The pci device structure
 *
 * returns The heci_device_device pointer on success, NULL on failure.
 */
struct iamt_heci_device *init_heci_device(struct pci_dev *pdev)
{
	struct iamt_heci_device *dev;

	dev = kzalloc(sizeof(struct iamt_heci_device), GFP_KERNEL);
	if (dev == NULL)
		return NULL;

	spin_lock_init(&dev->device_lock);
	spin_lock_init(&dev->crrdl_lock);
	spin_lock_init(&dev->rdcpl_lock);
	spin_lock_init(&dev->cmdl_lock);
	spin_lock_init(&dev->mecl_lock);
	spin_lock_init(&dev->conn_lock);
	spin_lock_init(&dev->hw_lock);
	spin_lock_init(&dev->write_lock);
	init_waitqueue_head(&dev->wait_recvd_msg);
	init_waitqueue_head(&dev->wait_stop_wd);
	init_waitqueue_head(&dev->wait_isr);
	atomic_set(&dev->recvd_isr, 0);
	atomic_set(&dev->recvd_msg, 0);
	atomic_set(&dev->recvd_wrr, 0);
	init_MUTEX(&dev->reinit_sem);
	atomic_set(&dev->heci_state, HECI_INITIALIZING);
	atomic_set(&dev->host_buffer_is_empty, 1);

	INIT_LIST_HEAD(&dev->conn_list);
	INIT_LIST_HEAD(&dev->file_list);

	heci_initialize_list(&dev->read_list);
	heci_initialize_list(&dev->read_cmpl_list);
	heci_initialize_list(&dev->ctrl_rd_list);
	heci_initialize_list(&dev->ctrl_wr_list);
	heci_initialize_list(&dev->write_list);
	heci_initialize_list(&dev->cmpl_list);
	heci_initialize_list(&dev->pthi_cmd_list);
	heci_initialize_list(&dev->heci_cmd_list);

	atomic_set(&dev->wd_stoped, 0);
	atomic_set(&dev->wd_bypass, 0);
	heci_init_heci_connection(&dev->wd_conn, NULL);
	atomic_inc(&dev->wd_conn.refcnt);
	heci_init_heci_connection(&dev->iamthif_conn, NULL);
	atomic_inc(&dev->iamthif_conn.refcnt);

	atomic_set(&dev->refcnt, 1);

	dev->pdev = pdev;
	return dev;
}

/**
 * put_heci_device - free unreferenced heci device structure
 *
 * @dev: Device object for our driver
 */
void put_heci_device(struct iamt_heci_device *dev)
{
	if (dev == NULL)
		return;

	if (!atomic_dec_and_test(&dev->refcnt))
		return;

	if (dev->mem_addr != NULL)
		iounmap(dev->mem_addr);
	kfree(dev);
}

/**
 * heci_wait_msg_timeout - wait for heci response
 *
 * @dev: Device object for our driver
 * @timeout: maximum wait timeout
 *
 * returns 0 if timeout, >0 if received, <0 on error
 */
static int heci_wait_msg_timeout(struct iamt_heci_device *dev, long timeout)
{
	return wait_event_interruptible_timeout(dev->wait_recvd_msg,
			(atomic_read(&dev->recvd_msg) != HECI_MSGINIT_NONE),
			 timeout);
}

/**
 * heci_hw_init  - init host and fw to start work.
 *
 * @dev: Device object for our driver
 *
 * returns 0 on success, <0 on failure.
 */
int heci_hw_init(struct iamt_heci_device *dev)
{
	int err = 0;

	spin_lock_bh(&dev->device_lock);
	spin_lock(&dev->hw_lock);
	dev->host_hw_state = read_heci_register(dev, H_CSR);
	dev->me_hw_state = read_heci_register(dev, ME_CSR_HA);
	DBG("host_hw_state = 0x%08x, mestate = 0x%08x.\n",
	    dev->host_hw_state, dev->me_hw_state);

	if ((dev->host_hw_state & H_IS) == H_IS) {
		/* acknowledge interrupt */
		heci_set_csr_register(dev);
	}
	spin_unlock(&dev->hw_lock);

	atomic_set(&dev->recvd_msg, HECI_MSGINIT_NONE);

	DBG("reset in start the heci device.\n");
	heci_reset(dev, 1);

	DBG("host_hw_state = 0x%08x, me_hw_state = 0x%08x.\n",
	    dev->host_hw_state, dev->me_hw_state);
	spin_unlock_bh(&dev->device_lock);

	/* wait for ME to turn on ME_RDY */
	if (atomic_read(&dev->recvd_msg) != HECI_MSGINIT_INTEROP)
		err = heci_wait_msg_timeout(dev, HECI_INTEROP_TIMEOUT);

	if (err == -ERESTARTSYS) {
		atomic_set(&dev->heci_state, HECI_DISABLED);
		atomic_set(&dev->recvd_msg, HECI_MSGINIT_NONE);
		DBG("wait_event_interruptible_timeout interrupted"
		    " on wait for ME to turn on ME_RDY.\n");
		return -EINTR;
	}
	if ((err == 0)
	    || (atomic_read(&dev->recvd_msg) != HECI_MSGINIT_INTEROP)) {
		if (err == 0)
			atomic_set(&dev->heci_state, HECI_DISABLED);
		atomic_set(&dev->recvd_msg, HECI_MSGINIT_NONE);
		DBG("wait_event_interruptible_timeout failed"
		    " on wait for ME to turn on ME_RDY.\n");
		return -ENODEV;
	}
	atomic_set(&dev->recvd_msg, HECI_MSGINIT_NONE);

	spin_lock_bh(&dev->hw_lock);
	dev->host_hw_state = read_heci_register(dev, H_CSR);
	dev->me_hw_state = read_heci_register(dev, ME_CSR_HA);
	if (((dev->host_hw_state & H_RDY) != H_RDY)
	      || ((dev->me_hw_state & ME_RDY_HRA) != ME_RDY_HRA)) {
		atomic_set(&dev->heci_state, HECI_DISABLED);
		DBG("host_hw_state = 0x%08x, me_hw_state = 0x%08x.\n",
		    dev->host_hw_state,
		    dev->me_hw_state);

		if ((dev->host_hw_state & H_RDY) != H_RDY)
			DBG("host turn off H_RDY.\n");

		if ((dev->me_hw_state & ME_RDY_HRA) != ME_RDY_HRA)
			DBG("ME turn off ME_RDY.\n");

		spin_unlock_bh(&dev->hw_lock);
		printk(KERN_ERR
		       "heci: link layer initialization failed.\n");
		return -ENODEV;
	}

	DBG("host_hw_state = 0x%08x, me_hw_state = 0x%08x.\n",
	    dev->host_hw_state, dev->me_hw_state);
	spin_unlock_bh(&dev->hw_lock);

	DBG("ME turn on ME_RDY and host turn on H_RDY.\n");
	printk(KERN_INFO "heci: link layer has been established.\n");
	return 0;
}

/**
 * heci_hw_reset  - reset fw via heci csr register.
 *
 * @dev: Device object for our driver
 * @interrupts: if interrupt should be enable after reset.
 *
 * NOTE: called under hw_lock
 */
static void heci_hw_reset(struct iamt_heci_device *dev, int interrupts)
{
	dev->host_hw_state |= (H_RST | H_IG);

	if (interrupts)
		heci_csr_enable_interrupts(dev);
	else
		heci_csr_disable_interrupts(dev);

	BUG_ON((dev->host_hw_state & H_RST) != H_RST);
	BUG_ON((dev->host_hw_state & H_RDY) != 0);
}

/**
 * heci_reset_io_list - clear list and free entries.
 *
 * @list:  An instance of our list structure
 */
static void heci_reset_io_list(struct heci_io_list *list)
{
	struct heci_cb_private *priv_cb_pos = NULL;
	struct heci_cb_private *priv_cb_next = NULL;

	if (list_empty(&list->cb_list))
		return;

	list_for_each_entry_safe(priv_cb_pos, priv_cb_next,
			&list->cb_list, cb_list) {
		list_del_init(&priv_cb_pos->cb_list);
		heci_free_io_cb(priv_cb_pos);
	}
}

/**
 * heci_reset_io  - resets IO operations.
 *
 * @dev: Device object for our driver
 *
 * NOTE: called under device_lock
 */
static void heci_reset_io(struct iamt_heci_device *dev)
{
	struct heci_connection *conn_pos = NULL;
	struct heci_connection *conn_next = NULL;

	spin_lock(&dev->conn_lock);
	list_for_each_entry_safe(conn_pos, conn_next, &dev->conn_list, link) {
		spin_lock(&conn_pos->file_lock);
		atomic_set(&conn_pos->state, HECI_FILE_DISCONNECTED);
		conn_pos->flow_ctrl_creds = 0;
		conn_pos->canceled = 1;
		conn_pos->read_cb = NULL;
		conn_pos->reading_state = HECI_IDLE;
		conn_pos->wr_state = HECI_IDLE;
		if (conn_pos == &dev->iamthif_conn)
			conn_pos->file = NULL;
		spin_unlock(&conn_pos->file_lock);
	}
	/* remove entry if already in list */
	DBG("list del iamthif and wd conn list.\n");
	list_del_init(&dev->wd_conn.link);
	list_del_init(&dev->iamthif_conn.link);
	spin_unlock(&dev->conn_lock);

	/* remove all waiting requests */
	heci_reset_io_list(&dev->write_list);
	spin_lock(&dev->crrdl_lock);
	heci_reset_io_list(&dev->read_list);
	spin_unlock(&dev->crrdl_lock);
}

/**
 * heci_reset  - reset host and fw.
 *
 * @dev: Device object for our driver
 * @interrupts: if interrupt should be enable after reset.
 *
 * NOTE: called under device_lock
 */
void heci_reset(struct iamt_heci_device *dev, int interrupts)
{
	struct heci_file_priv *file_pos = NULL;
	struct heci_file_priv *file_next = NULL;
	int unexpected = 0;
	enum heci_states heci_state;

	heci_state = atomic_read(&dev->heci_state);
	if (heci_state == HECI_RECOVERING_FROM_RESET) {
		dev->need_reset = 1;
		return;
	}
	if ((heci_state != HECI_INITIALIZING) &&
	    (heci_state != HECI_DISABLED) &&
	    (heci_state != HECI_POWER_DOWN) &&
	    (heci_state != HECI_REMOVING)) {
		if (heci_state != HECI_POWER_UP)
			unexpected = 1;
		atomic_set(&dev->heci_state, HECI_RESETING);
	}

	if (dev->reinit_tsk != NULL) {
		set_tsk_thread_flag(dev->reinit_tsk, TIF_SIGPENDING);
		wake_up_process(dev->reinit_tsk);
	}

	spin_lock(&dev->hw_lock);
	dev->host_hw_state = read_heci_register(dev, H_CSR);
	DBG("before reset host_hw_state = 0x%08x.\n",
	    dev->host_hw_state);

	heci_hw_reset(dev, interrupts);

	dev->host_hw_state &= ~H_RST;
	dev->host_hw_state |= H_IG;
	write_heci_register(dev, H_CSR, dev->host_hw_state);
	DBG("currently saved host_hw_state = 0x%08x.\n",
	    dev->host_hw_state);
	spin_unlock(&dev->hw_lock);

	dev->need_reset = 0;

	if (atomic_read(&dev->heci_state) != HECI_INITIALIZING)
		heci_reset_io(dev);

	spin_lock(&dev->mecl_lock);
	dev->num_heci_me_clients = 0;
	spin_unlock(&dev->mecl_lock);

	dev->rd_msg_hdr = 0;
	dev->wd_stop = 0;
	dev->wd_pending = 0;
	dev->wd_due_counter = 0;
	dev->extra_write_index = 0;

	spin_lock(&dev->hw_lock);
	/* update the state of the registers after reset */
	dev->host_hw_state = read_heci_register(dev, H_CSR);
	dev->me_hw_state = read_heci_register(dev, ME_CSR_HA);
	DBG("after reset host_hw_state = 0x%08x, me_hw_state = 0x%08x.\n",
	    dev->host_hw_state, dev->me_hw_state);
	spin_unlock(&dev->hw_lock);

	if (unexpected)
		printk(KERN_WARNING "heci: unexpected reset.\n");

	spin_lock(&dev->conn_lock);
	/* Wake up all readings and writings so they can be interrupted */
	list_for_each_entry_safe(file_pos, file_next, &dev->file_list, link) {
		if (waitqueue_active(&file_pos->rx_wait)) {
			printk(KERN_INFO "heci: waking up client\n");
			wake_up_interruptible(&file_pos->rx_wait);
		}
		if (waitqueue_active(&file_pos->tx_wait)) {
			printk(KERN_INFO "heci: waking up writing client\n");
			wake_up_interruptible(&file_pos->tx_wait);
		}
	}
	spin_unlock(&dev->conn_lock);
}

/**
 * heci_initialize_clients  -  initializes heci communication.
 *
 * @dev: Device object for our driver
 *
 * returns 0 on success, !=0 on failure
 */
int heci_initialize_clients(struct iamt_heci_device *dev)
{
	int ret;

	msleep(100); /* FW needs time to be ready to talk with us */
	DBG("link is established start sending messages.\n");
	/* link is established start sending messages. */
	ret = host_start_message(dev);
	if (ret != 0) {
		DBG("start sending messages failed.\n");
		return ret;
	}
	/* enumerate clients */

	ret = host_enum_clients_message(dev);
	if (ret != 0) {
		DBG("enum clients failed.\n");
		return ret;
	}
	/* allocate storage for ME clients representation */
	ret = allocate_me_clients_storage(dev);
	if (ret != 0) {
		DBG("allocate clients failed.\n");
		return ret;
	}

	heci_check_asf_mode(dev);
	/* heci initialization wd */
	ret = host_init_wd(dev);
	if (ret == -ERESTARTSYS)
		return -EINTR;
	/* heci initialization iamthif client */
	ret = host_init_iamthif(dev);
	if (ret == -ERESTARTSYS)
		return -EINTR;

	spin_lock_bh(&dev->device_lock);
	if (dev->need_reset == 1) {
		dev->need_reset = 0;
		atomic_set(&dev->heci_state, HECI_DISABLED);
		spin_unlock_bh(&dev->device_lock);
		return -ENODEV;
	}

	spin_lock(&dev->conn_lock);
	memset(dev->heci_host_clients, 0, sizeof(dev->heci_host_clients));
	dev->open_handle_count = 0;
	dev->heci_host_clients[0] |= 7;
	dev->current_host_client_id = HECI_LAST_FIX_HOST_CLIENT_ID;
	spin_unlock(&dev->conn_lock);

	atomic_set(&dev->heci_state, HECI_ENABLED);
	spin_unlock_bh(&dev->device_lock);
	DBG("initialization heci clients successful.\n");
	return 0;
}


#ifndef HECI_TASK_NONFREEZABLE
#define HECI_TASK_NONFREEZABLE
#endif

/**
 * heci_task_initialize_clients  -  heci reinitialization task.
 *
 * @data: Device object for our driver
 *
 * returns 0 on success, !=0 on failure
 */
int heci_task_initialize_clients(void *data)
{
	int ret;
	struct iamt_heci_device *dev = (struct iamt_heci_device *) data;

	if (down_trylock(&dev->reinit_sem)) {
		DBG("reinit task already started.\n");
		return 0;
	}

	spin_lock_bh(&dev->device_lock);
	dev->reinit_tsk = current;
	spin_unlock_bh(&dev->device_lock);

	HECI_TASK_NONFREEZABLE;

	ret = heci_initialize_clients(dev);

	spin_lock_bh(&dev->device_lock);
	dev->reinit_tsk = NULL;
	spin_unlock_bh(&dev->device_lock);

	up(&dev->reinit_sem);

	return ret;
}

/**
 * host_start_message - heci host send start message.
 *
 * @dev: Device object for our driver
 *
 * returns 0 on success, <0 on failure.
 */
static int host_start_message(struct iamt_heci_device *dev)
{
	long timeout = 60;	/* 60 second */

	struct heci_msg_hdr *heci_hdr;
	struct hbm_host_version_request *host_start_req;
	struct hbm_host_stop_request *host_stop_req;
	int err = 0;

	spin_lock_bh(&dev->write_lock);
	/* host start message */
	heci_hdr = (struct heci_msg_hdr *) &dev->wr_msg_buf[0];
	heci_hdr->host_addr = 0;
	heci_hdr->me_addr = 0;
	heci_hdr->length = sizeof(struct hbm_host_version_request);
	heci_hdr->msg_complete = 1;
	heci_hdr->reserved = 0;

	host_start_req =
	    (struct hbm_host_version_request *) &dev->wr_msg_buf[1];
	memset(host_start_req, 0, sizeof(struct hbm_host_version_request));
	host_start_req->cmd.cmd = HOST_START_REQ_CMD;
	host_start_req->host_version.major_version = HBM_MAJOR_VERSION;
	host_start_req->host_version.minor_version = HBM_MINOR_VERSION;

	atomic_set(&dev->recvd_msg, HECI_MSGINIT_NONE);
	if (!heci_write_message(dev, heci_hdr,
				(unsigned char *) (host_start_req),
				heci_hdr->length)) {
		atomic_set(&dev->heci_state, HECI_DISABLED);
		spin_unlock_bh(&dev->write_lock);
		DBG("send version to fw fail.\n");
		return -ENODEV;
	}
	spin_unlock_bh(&dev->write_lock);

	DBG("call wait_event_interruptible_timeout for response message.\n");
	/* wait for response */
	err = heci_wait_msg_timeout(dev, timeout * HZ);
	if (err == -ERESTARTSYS) {
		atomic_set(&dev->heci_state, HECI_DISABLED);
		atomic_set(&dev->recvd_msg, HECI_MSGINIT_NONE);
		DBG("wait_event_interruptible_timeout interrupted"
		    " on host start response message.\n");
		return -EINTR;
	}
	if ((err == 0)
	    || (atomic_read(&dev->recvd_msg) != HECI_MSGINIT_START)) {
		if (err == 0)
			atomic_set(&dev->heci_state, HECI_DISABLED);
		atomic_set(&dev->recvd_msg, HECI_MSGINIT_NONE);
		DBG("wait_event_interruptible_timeout failed"
		    " on host start response message.\n");
		return -ENODEV;
	}
	atomic_set(&dev->recvd_msg, HECI_MSGINIT_NONE);
	DBG("wait_timeout successful on host start response message.\n");

	spin_lock_bh(&dev->write_lock);
	if ((dev->version.major_version != HBM_MAJOR_VERSION) ||
	    (dev->version.minor_version != HBM_MINOR_VERSION)) {
		/* send stop message */
		heci_hdr->host_addr = 0;
		heci_hdr->me_addr = 0;
		heci_hdr->length = sizeof(struct hbm_host_stop_request);
		heci_hdr->msg_complete = 1;
		heci_hdr->reserved = 0;

		host_stop_req =
		    (struct hbm_host_stop_request *) &dev->wr_msg_buf[1];

		memset(host_stop_req, 0, sizeof(struct hbm_host_stop_request));
		host_stop_req->cmd.cmd = HOST_STOP_REQ_CMD;
		host_stop_req->reason = DRIVER_STOP_REQUEST;
		heci_write_message(dev, heci_hdr,
				   (unsigned char *) (host_stop_req),
				   heci_hdr->length);
		atomic_set(&dev->heci_state, HECI_DISABLED);
		spin_unlock_bh(&dev->write_lock);
		DBG("version mismatch.\n");
		return -ENODEV;
	}
	spin_unlock_bh(&dev->write_lock);

	return 0;
}

/**
 * host_enum_clients_message - host send enumeration client request message.
 *
 * @dev: Device object for our driver
 *
 * returns 0 on success, <0 on failure.
 */
static int host_enum_clients_message(struct iamt_heci_device *dev)
{
	long timeout = 5;	/* 5 second */
	struct heci_msg_hdr *heci_hdr;
	struct hbm_host_enum_request *host_enum_req;
	int err = 0;
	__u8 i, j;

	spin_lock_bh(&dev->write_lock);
	heci_hdr = (struct heci_msg_hdr *) &dev->wr_msg_buf[0];
	/* enumerate clients */
	heci_hdr->host_addr = 0;
	heci_hdr->me_addr = 0;
	heci_hdr->length = sizeof(struct hbm_host_enum_request);
	heci_hdr->msg_complete = 1;
	heci_hdr->reserved = 0;

	host_enum_req = (struct hbm_host_enum_request *) &dev->wr_msg_buf[1];
	memset(host_enum_req, 0, sizeof(struct hbm_host_enum_request));
	host_enum_req->cmd.cmd = HOST_ENUM_REQ_CMD;

	atomic_set(&dev->recvd_msg, HECI_MSGINIT_NONE);
	if (!heci_write_message(dev, heci_hdr,
				(unsigned char *) (host_enum_req),
				heci_hdr->length)) {
		atomic_set(&dev->heci_state, HECI_DISABLED);
		spin_unlock_bh(&dev->write_lock);
		DBG("send enumeration request failed.\n");
		return -ENODEV;
	}
	spin_unlock_bh(&dev->write_lock);

	/* wait for response */
	err = heci_wait_msg_timeout(dev, timeout * HZ);
	if (err == -ERESTARTSYS) {
		atomic_set(&dev->heci_state, HECI_DISABLED);
		atomic_set(&dev->recvd_msg, HECI_MSGINIT_NONE);
		DBG("wait_event_interruptible_timeout interrupted"
		    " on enumeration clients response message.\n");
		return -EINTR;
	}
	if ((err == 0)
	    || (atomic_read(&dev->recvd_msg) != HECI_MSGINIT_ENUM_CLIENTS)) {
		if (err == 0)
			atomic_set(&dev->heci_state, HECI_DISABLED);
		atomic_set(&dev->recvd_msg, HECI_MSGINIT_NONE);
		DBG("wait_event_interruptible_timeout failed"
		    " on enumeration clients response message.\n");
		return -ENODEV;
	}
	atomic_set(&dev->recvd_msg, HECI_MSGINIT_NONE);

	spin_lock_bh(&dev->mecl_lock);
	/* count how many ME clients we have */
	for (i = 0; i < sizeof(dev->heci_me_clients); i++) {
		for (j = 0; j < 8; j++) {
			if ((dev->heci_me_clients[i] & (1 << j)) != 0)
				dev->num_heci_me_clients++;

		}
	}
	spin_unlock_bh(&dev->mecl_lock);

	return 0;
}

/**
 * host_client_properties - reads properties for client
 *
 * @dev: Device object for our driver
 * @idx: client index in me client array
 * @client_id: id of the client
 *
 * returns 0 on success, <0 on failure.
 */
static int host_client_properties(struct iamt_heci_device *dev,
				  struct heci_me_client *client)
{
	long timeout = 10;	/* 10 second */
	struct heci_msg_hdr *heci_hdr;
	struct hbm_props_request *host_cli_req;
	int err;

	spin_lock_bh(&dev->write_lock);
	heci_hdr = (struct heci_msg_hdr *) &dev->wr_msg_buf[0];
	heci_hdr->host_addr = 0;
	heci_hdr->me_addr = 0;
	heci_hdr->length = sizeof(struct hbm_props_request);
	heci_hdr->msg_complete = 1;
	heci_hdr->reserved = 0;

	host_cli_req = (struct hbm_props_request *) &dev->wr_msg_buf[1];
	memset(host_cli_req, 0, sizeof(struct hbm_props_request));
	host_cli_req->cmd.cmd = HOST_CLIENT_PROPERTEIS_REQ_CMD;
	host_cli_req->address = client->client_id;

	atomic_set(&dev->recvd_msg, HECI_MSGINIT_NONE);
	if (!heci_write_message(dev, heci_hdr,
				(unsigned char *) (host_cli_req),
				heci_hdr->length)) {
		atomic_set(&dev->heci_state, HECI_DISABLED);
		spin_unlock_bh(&dev->write_lock);
		DBG("send props request failed.\n");
		return -ENODEV;
	}
	spin_unlock_bh(&dev->write_lock);

	/* wait for response */
	err = heci_wait_msg_timeout(dev, timeout * HZ);
	if (err == -ERESTARTSYS) {
		atomic_set(&dev->heci_state, HECI_DISABLED);
		atomic_set(&dev->recvd_msg, HECI_MSGINIT_NONE);
		DBG("wait_event_interruptible_timeout interrupted"
		    " on props resp msg.\n");
		return -EINTR;
	}
	if ((err == 0)
	    || (atomic_read(&dev->recvd_msg) != HECI_MSGINIT_CLIENT_PROPS)) {
		if (err == 0)
			atomic_set(&dev->heci_state, HECI_DISABLED);
		atomic_set(&dev->recvd_msg, HECI_MSGINIT_NONE);
		DBG("wait_event_interruptible_timeout failed"
		    " on props resp msg.\n");
		return -ENODEV;
	}
	atomic_set(&dev->recvd_msg, HECI_MSGINIT_NONE);

	return 0;
}

/**
 * allocate_me_clients_storage - allocate storage for me clients
 *
 * @dev: Device object for our driver
 *
 * returns 0 on success, <0 on failure.
 */
static int allocate_me_clients_storage(struct iamt_heci_device *dev)
{
	struct heci_me_client *clients;
	struct heci_me_client *client;
	__u8 num, i, j;
	int err;

	spin_lock_bh(&dev->mecl_lock);
	if (dev->num_heci_me_clients <= 0) {
		kfree(dev->me_clients);
		dev->me_clients = NULL;
		dev->num_heci_me_clients = 0;
		atomic_set(&dev->heci_state, HECI_DISABLED);
		spin_unlock_bh(&dev->mecl_lock);
		return 0;
	}
	spin_unlock_bh(&dev->mecl_lock);

	/* allocate storage for ME clients representation */
	clients = kcalloc(dev->num_heci_me_clients,
			sizeof(struct heci_me_client), GFP_KERNEL);

	spin_lock_bh(&dev->mecl_lock);
	kfree(dev->me_clients);
	dev->me_clients = clients;
	if (clients == NULL) {
		dev->num_heci_me_clients = 0;
		atomic_set(&dev->heci_state, HECI_DISABLED);
		spin_unlock_bh(&dev->mecl_lock);
		DBG("memory allocation for ME clients failed.\n");
		return -ENOMEM;
	}

	num = 0;
	for (i = 0; i < sizeof(dev->heci_me_clients); i++) {
		for (j = 0; j < 8; j++) {
			if ((dev->heci_me_clients[i] & (1 << j)) != 0) {
				client = &dev->me_clients[num];
				client->client_id = (i * 8) + j;
				client->flow_ctrl_creds = 0;
				spin_unlock_bh(&dev->mecl_lock);
				err = host_client_properties(dev, client);
				spin_lock_bh(&dev->mecl_lock);
				if (err != 0) {
					kfree(dev->me_clients);
					dev->me_clients = NULL;
					dev->num_heci_me_clients = 0;
					spin_unlock_bh(&dev->mecl_lock);
					return err;
				}
				num++;
			}
		}
	}
	spin_unlock_bh(&dev->mecl_lock);
	return 0;
}

/**
 * heci_find_me_client - search for ME client guid
 *                       sets client_id in heci_connection if found
 *
 * @dev: Device object for our driver
 * @conn: connection data of the file object to set client_id in
 * @cguid: searched guid of ME client
 * @client_id: id of host client to be set in file private structure
 *
 * returns ME client index
 */
static __u8 heci_find_me_client(struct iamt_heci_device *dev,
				struct heci_connection *conn,
				const struct guid *cguid, __u8 client_id)
{
	__u8 i;

	if ((dev == NULL) || (conn == NULL) || (cguid == NULL))
		return 0;

	for (i = 0; i < dev->num_heci_me_clients; i++) {
		if (memcmp(cguid,
			   &dev->me_clients[i].props.protocol_name,
			   sizeof(struct guid)) == 0) {
			conn->me_client_id = dev->me_clients[i].client_id;
			conn->max_msg_length =
				    dev->me_clients[i].props.max_msg_length;
			atomic_set(&conn->state, HECI_FILE_CONNECTING);
			conn->host_client_id = client_id;

			spin_lock(&dev->conn_lock);
			list_add_tail(&conn->link, &dev->conn_list);
			spin_unlock(&dev->conn_lock);
			return i;
		}
	}
	return 0;
}

/**
 * heci_check_asf_mode - check for ASF client
 *
 * @dev: Device object for our driver
 */
static void heci_check_asf_mode(struct iamt_heci_device *dev)
{
	__u8 i;

	spin_lock_bh(&dev->mecl_lock);
	dev->asf_mode = 0;
	/* find ME ASF client - otherwise assume AMT mode */
	DBG("find ME ASF client - otherwise assume AMT mode.\n");
	for (i = 0; i < dev->num_heci_me_clients; i++) {
		if (memcmp(&heci_asf_guid,
				&dev->me_clients[i].props.protocol_name,
				sizeof(struct guid)) == 0) {
			dev->asf_mode = 1;
			spin_unlock_bh(&dev->mecl_lock);
			DBG("found ME ASF client.\n");
			return;
		}
	}
	spin_unlock_bh(&dev->mecl_lock);
	DBG("assume AMT mode.\n");
}

/**
 * heci_connect_me_client - connect ME client
 *
 * @dev: Device object for our driver
 * @conn: connection data of the file object
 * @timeout: connect timeout in seconds
 *
 * returns 0 - if connected, !=0 - if not
 */
static int heci_connect_me_client(struct iamt_heci_device *dev,
				   struct heci_connection *conn,
				   long timeout)
{
	int err = 0;

	if ((dev == NULL) || (conn == NULL))
		return -ENODEV;

	if (!heci_connect(dev, conn)) {
		DBG("failed to call heci_connect for client_id=%d.\n",
		    conn->host_client_id);
		spin_lock_bh(&dev->conn_lock);
		list_del_init(&conn->link);
		spin_unlock_bh(&dev->conn_lock);
		atomic_set(&conn->state, HECI_FILE_DISCONNECTED);
		return -ENODEV;
	}

	err = wait_event_interruptible_timeout(conn->cc_wait,
		    ((atomic_read(&conn->state) == HECI_FILE_CONNECTED) ||
		     (atomic_read(&conn->state) == HECI_FILE_DISCONNECTED)),
		    timeout * HZ);
	if (atomic_read(&conn->state) != HECI_FILE_CONNECTED) {
		spin_lock_bh(&dev->conn_lock);
		list_del_init(&conn->link);
		spin_unlock_bh(&dev->conn_lock);
		DBG("failed to connect client_id=%d state=%d.\n",
		    conn->host_client_id, atomic_read(&conn->state));
		atomic_set(&conn->state, HECI_FILE_DISCONNECTED);
		if (err) {
			DBG("failed connect err=%08x\n", err);
			return err;
		}
		return -ETIME;
	}
	DBG("successfully connected client_id=%d.\n",
	    conn->host_client_id);
	spin_lock_bh(&conn->file_lock);
	conn->reading_state = HECI_IDLE;
	conn->wr_state = HECI_IDLE;
	spin_unlock_bh(&conn->file_lock);
	return 0;
}

/**
 * host_init_wd - heci watchdog initialization
 *
 * @dev: Device object for our driver
 *
 * returns 0 - if connected, !=0 - if not
 */
static int host_init_wd(struct iamt_heci_device *dev)
{
	int ret = 0;
	struct heci_connection *conn = &dev->wd_conn;

	spin_lock_bh(&conn->file_lock);
	heci_reset_heci_connection(conn, NULL);
	atomic_inc(&conn->refcnt);
	atomic_set(&conn->state, HECI_FILE_DISCONNECTED);

	/* look for WD client and connect to it */
	dev->wd_timeout = 0;

	memset(dev->wd_data, 0, HECI_START_WD_DATA_SIZE);
	if (dev->asf_mode) {
		memcpy(dev->wd_data, heci_stop_wd_params, HECI_WD_PARAMS_SIZE);
	} else {
		/* AMT mode */
		dev->wd_timeout = AMT_WD_VALUE;
		DBG("dev->wd_timeout=%d.\n", dev->wd_timeout);
		memcpy(dev->wd_data, heci_start_wd_params, HECI_WD_PARAMS_SIZE);
		memcpy(dev->wd_data + HECI_WD_PARAMS_SIZE,
		       &dev->wd_timeout, sizeof(__u16));
	}

	spin_lock(&dev->mecl_lock);
	/* find ME WD client */
	heci_find_me_client(dev, conn, &heci_wd_guid, HECI_WD_HOST_CLIENT_ID);
	spin_unlock(&dev->mecl_lock);
	DBG("check wd_conn\n");
	if (atomic_read(&conn->state) == HECI_FILE_CONNECTING) {
		spin_unlock_bh(&conn->file_lock);
		ret = heci_connect_me_client(dev, conn, HECI_CONNECT_TIMEOUT);
		spin_lock_bh(&conn->file_lock);
		if (ret == 0) {
			DBG("dev->wd_timeout=%d.\n", dev->wd_timeout);
			if (dev->wd_timeout != 0)
				dev->wd_due_counter = 1;
			else
				dev->wd_due_counter = 0;
			DBG("successfully connected to WD client.\n");
		}
	} else
		DBG("failed to find WD client.\n");

	dev->wd_timer.function = &heci_wd_timer;
	dev->wd_timer.data = (unsigned long) dev;
	spin_unlock_bh(&conn->file_lock);
	return ret;
}

/**
 * host_init_iamthif - heci iamthif client initialization
 *
 * @dev: Device object for our driver
 *
 * returns 0 - if connected, !=0 - if not
 */
static int host_init_iamthif(struct iamt_heci_device *dev)
{
	__u8 i;
	int ret = 0;
	struct heci_connection *conn = &dev->iamthif_conn;

	spin_lock_bh(&conn->file_lock);
	heci_reset_heci_connection(conn, NULL);
	atomic_inc(&conn->refcnt);
	atomic_set(&conn->state, HECI_FILE_DISCONNECTED);

	spin_lock(&dev->mecl_lock);
	/* find ME PTHI client */
	i = heci_find_me_client(dev, conn, &heci_pthi_guid,
				HECI_IAMTHIF_HOST_CLIENT_ID);
	if (atomic_read(&conn->state) != HECI_FILE_CONNECTING) {
		DBG("failed to find iamthif client.\n");
		spin_unlock(&dev->mecl_lock);
		return 0;
	}
	BUG_ON(dev->me_clients[i].props.max_msg_length != IAMTHIF_MTU);
	spin_unlock(&dev->mecl_lock);
	spin_unlock_bh(&conn->file_lock);

	ret = heci_connect_me_client(dev, conn, HECI_CONNECT_TIMEOUT);
	if (ret == 0)
		DBG("connected to iamthif client.\n");
	return ret;
}

