/*
 * Part of Intel(R) Manageability Engine Interface Linux driver
 *
 * Copyright (c) 2003 - 2008 Intel Corp.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    substantially similar to the "NO WARRANTY" disclaimer below
 *    ("Disclaimer") and any redistribution must be conditioned upon
 *    including a substantially similar Disclaimer requirement for further
 *    binary redistribution.
 * 3. Neither the names of the above-listed copyright holders nor the names
 *    of any contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * NO WARRANTY
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 */

#include "heci.h"
#include "heci_version.h"

/**
 * heci_ioctl_get_version - the get driver version IOCTL function
 *
 * @dev: Device object for our driver
 * @u_msg: pointer to user data struct in user space
 * @k_msg: data in kernel on the stack
 *
 * returns 0 on success, <0 on failure.
 */
int heci_ioctl_get_version(struct iamt_heci_device *dev,
			   struct heci_message_data *u_msg,
			   struct heci_message_data k_msg)
{
	int rets = 0;
	struct heci_driver_version *version;
	struct heci_message_data res_msg;

	if (dev == NULL)
		return -ENODEV;

	if (k_msg.size < (sizeof(struct heci_driver_version) - 2)) {
		DBG("user buffer less than heci_driver_version.\n");
		return -EMSGSIZE;
	}

	res_msg.data = kmalloc(sizeof(struct heci_driver_version), GFP_KERNEL);
	if (res_msg.data == NULL) {
		DBG("failed allocation response buffer size = %d.\n",
		    (int) sizeof(struct heci_driver_version));
		return -ENOMEM;
	}

	version = (struct heci_driver_version *) res_msg.data;
	version->major = MAJOR_VERSION;
	version->minor = MINOR_VERSION;
	version->hotfix = QUICK_FIX_NUMBER;
	version->build = VER_BUILD;
	res_msg.size = sizeof(struct heci_driver_version);
	if (k_msg.size < sizeof(struct heci_driver_version))
		res_msg.size -= 2;

	/* now copy the data to user space */
	if (copy_to_user(k_msg.data, res_msg.data, res_msg.size)) {
		rets = -EFAULT;
		goto end;
	}
	if (put_user(res_msg.size, &u_msg->size)) {
		rets = -EFAULT;
		goto end;
	}
end:
	kfree(res_msg.data);
	return rets;
}

/**
 * heci_ioctl_connect_client - the connect to fw client IOCTL function
 *
 * @dev: Device object for our driver
 * @u_msg: pointer to user data struct in user space
 * @k_msg: data in kernel on the stack
 * @file_ext: private data of the file object
 *
 * returns 0 on success, <0 on failure.
 */
int heci_ioctl_connect_client(struct iamt_heci_device *dev,
			      struct heci_message_data *u_msg,
			      struct heci_message_data k_msg,
			      struct heci_file_priv *file_ext)
{
	int rets = 0;
	struct heci_message_data *req_msg = NULL;
	struct heci_cb_private *priv_cb = NULL;
	struct heci_client client;
	struct heci_connection *conn;
	long timeout = HECI_CONNECT_TIMEOUT; /* in second */
	__u8 i;
	int err = 0;

	DBG("entry\n");
	if ((dev == NULL) || (file_ext == NULL))
		return -ENODEV;

	if (file_ext->conn == NULL)
		return -ENODEV;

	if (k_msg.size != sizeof(struct guid)) {
		DBG("user buffer size is not equal to size of struct "
				"guid(16).\n");
		return -EMSGSIZE;
	}

	if (k_msg.data == NULL)
		return -EIO;

	if (file_ext->conn == &dev->iamthif_conn) {
		conn = heci_alloc_heci_connection(file_ext->file);
		if (conn == NULL)
			return -ENOMEM;
		conn->host_client_id = file_ext->host_client_id;
		atomic_set(&conn->state, HECI_FILE_INITIALIZING);
	} else
		conn = file_ext->conn;

	priv_cb = heci_init_io_cb(conn, HECI_IOCTL, file_ext,
					sizeof(struct guid), 0);
	if (priv_cb == NULL) {
		if ((file_ext->conn == &dev->iamthif_conn) &&
		    (conn != file_ext->conn))
			kfree(conn);
		return -ENOMEM;
	}
	req_msg = &priv_cb->request_buffer;

	/* copy the message to kernel space -
	 * use a pointer already copied into kernel space
	 */
	if (copy_from_user(req_msg->data, k_msg.data, sizeof(struct guid))) {
		rets = -EFAULT;
		goto free;
	}

	if (atomic_read(&dev->heci_state) != HECI_ENABLED) {
		rets = -ENODEV;
		goto free;
	}
	if ((atomic_read(&file_ext->conn->state) != HECI_FILE_INITIALIZING) &&
	    (atomic_read(&file_ext->conn->state) != HECI_FILE_DISCONNECTED)) {
		rets = -EBUSY;
		goto free;
	}

	spin_lock_bh(&conn->file_lock);
	spin_lock(&dev->mecl_lock);
	/* find ME client we're trying to connect to */
	for (i = 0; i < dev->num_heci_me_clients; i++) {
		if (memcmp((struct guid *)req_msg->data,
			    &dev->me_clients[i].props.protocol_name,
			    sizeof(struct guid)) == 0)
			break;
	}
	if (i == dev->num_heci_me_clients) {
		spin_unlock(&dev->mecl_lock);
		spin_unlock_bh(&conn->file_lock);
		rets = -ENODEV;
		goto free;
	}

	/* prepare the output buffer */
	client.max_msg_length = dev->me_clients[i].props.max_msg_length;
	client.protocol_version = dev->me_clients[i].props.protocol_version;
	conn->max_msg_length = dev->me_clients[i].props.max_msg_length;
	if (dev->me_clients[i].props.fixed_address == 0) {
		conn->me_client_id = dev->me_clients[i].client_id;
		atomic_set(&conn->state, HECI_FILE_CONNECTING);
	}
	spin_unlock(&dev->mecl_lock);
	spin_unlock_bh(&conn->file_lock);

	if (file_ext->conn == &dev->iamthif_conn) {
		atomic_inc(&conn->refcnt);
		file_ext->conn = conn;
		spin_lock_bh(&dev->conn_lock);
		list_add_tail(&conn->link, &dev->conn_list);
		spin_unlock_bh(&dev->conn_lock);
	}

	/* if we're connecting to PTHI client so we will use the existing
	 * connection
	 */
	if (memcmp((struct guid *)req_msg->data, &heci_pthi_guid,
				sizeof(struct guid)) == 0) {
		if (atomic_read(&dev->iamthif_conn.state)
				!= HECI_FILE_CONNECTED) {
			atomic_set(&conn->state, HECI_FILE_DISCONNECTED);
			rets = -ENODEV;
			goto free;
		}
		spin_lock_bh(&dev->conn_lock);
		list_del_init(&conn->link);
		spin_unlock_bh(&dev->conn_lock);

		DBG("free file connection data memory.\n");
		heci_free_heci_connection(conn);
		file_ext->conn = &dev->iamthif_conn;
		conn = file_ext->conn;
		heci_free_io_cb(priv_cb);
		priv_cb = NULL;
		goto copy;
	}


	if (atomic_read(&conn->state) != HECI_FILE_CONNECTING) {
		rets = -ENODEV;
		goto free;
	}

	spin_lock_bh(&dev->cmdl_lock);
	list_add_tail(&priv_cb->cb_list, &dev->heci_cmd_list.cb_list);
	spin_unlock_bh(&dev->cmdl_lock);
	atomic_inc(&dev->recvd_wrr);
	wake_up_interruptible(&dev->wait_isr);

	err = wait_event_interruptible_timeout(conn->cc_wait,
			((atomic_read(&conn->state) == HECI_FILE_CONNECTED)
			 || (atomic_read(&conn->state)
				== HECI_FILE_DISCONNECTED)),
			timeout * HZ);

	if (atomic_read(&conn->state) != HECI_FILE_CONNECTED) {
		DBG("failed to connect to FW client.conn->state = %d.\n",
		    atomic_read(&conn->state));
		if (err == 0) {
			DBG("wait_event_interruptible_timeout failed on client"
			    " connect message fw response message.\n");
		}
		rets = -EFAULT;
		if (err == -ERESTARTSYS)
			rets = -EINTR;
		if (priv_cb != NULL) {
			err = heci_flush_ctrl_lists(dev, conn, priv_cb);
			if (err != 0)
				heci_free_io_cb(priv_cb);
			priv_cb = NULL;
		}
		return rets;
	}
	priv_cb = NULL;

copy:
	DBG("successfully connected to FW client.\n");
	rets = conn->status;
	/* now copy the data to user space */
	if (copy_to_user(k_msg.data, &client, sizeof(struct heci_client)))
		return -EFAULT;
	if (put_user(sizeof(struct heci_client), &u_msg->size))
		return -EFAULT;

	DBG("return %d\n", rets);
	return rets;

free:
	DBG("free connect cb memory.");
	heci_free_io_cb(priv_cb);
	DBG("return %d\n", rets);
	return rets;
}

/**
 * heci_ioctl_wd  - the wd IOCTL function
 *
 * @dev: Device object for our driver
 * @k_msg: data in kernel on the stack
 *
 * returns 0 on success, <0 on failure.
 */
int heci_ioctl_wd(struct iamt_heci_device *dev,
		  struct heci_message_data k_msg)
{
	int rets = 0;
	struct heci_message_data req_msg; /* in kernel on the stack */

	DBG("entry\n");
	if (k_msg.size != HECI_WATCHDOG_DATA_SIZE) {
		DBG("user buffer has invalid size.\n");
		return -EMSGSIZE;
	}

	req_msg.data = kmalloc(HECI_WATCHDOG_DATA_SIZE, GFP_KERNEL);
	if (req_msg.data == NULL) {
		DBG("failed allocation request buffer size = %d.\n",
		    HECI_WATCHDOG_DATA_SIZE);
		return -ENOMEM;
	}
	req_msg.size = HECI_WATCHDOG_DATA_SIZE;

	/* copy the message to kernel space - use a pointer already
	 * copied into kernel space
	 */
	if (copy_from_user(req_msg.data, k_msg.data, req_msg.size)) {
		rets = -EFAULT;
		goto end;
	}

	spin_lock_bh(&dev->wd_conn.file_lock);
	if ((atomic_read(&dev->heci_state) != HECI_ENABLED)
	    || (atomic_read(&dev->wd_conn.state) != HECI_FILE_CONNECTED)) {
		spin_unlock_bh(&dev->wd_conn.file_lock);
		rets = -ENODEV;
		goto end;
	}

	if (!dev->asf_mode) {
		spin_unlock_bh(&dev->wd_conn.file_lock);
		rets = -EIO;
		goto end;
	}

	memcpy(&dev->wd_data[HECI_WD_PARAMS_SIZE], req_msg.data,
	       HECI_WATCHDOG_DATA_SIZE);

	dev->wd_timeout = (req_msg.data[1] << 8) + req_msg.data[0];
	dev->wd_pending = 0;
	dev->wd_due_counter = 1;	/* next timer */
	if (dev->wd_timeout == 0)
		memcpy(dev->wd_data, heci_stop_wd_params, HECI_WD_PARAMS_SIZE);
	else {
		memcpy(dev->wd_data, heci_start_wd_params, HECI_WD_PARAMS_SIZE);
		mod_timer(&dev->wd_timer, jiffies);
	}
	spin_unlock_bh(&dev->wd_conn.file_lock);
end:
	kfree(req_msg.data);
	DBG("return %d\n", rets);
	return rets;
}


/**
 * heci_ioctl_bypass_wd  - the bypass_wd IOCTL function
 *
 * @dev: Device object for our driver
 * @k_msg: data in kernel on the stack
 *
 * returns 0 on success, <0 on failure.
 */
int heci_ioctl_bypass_wd(struct iamt_heci_device *dev,
		  struct heci_message_data k_msg)
{
	__u8 flag = 0;

	DBG("entry\n");
	if (k_msg.size < 1) {
		DBG("user buffer less than HECI_WATCHDOG_DATA_SIZE.\n");
		return -EMSGSIZE;
	}

	if (copy_from_user(&flag, k_msg.data, 1))
		return -EFAULT;

	atomic_set(&dev->wd_bypass, ((flag) ? 1 : 0));

	return 0;
}

/**
 * heci_find_list_file_entry - finds a list entry for current file
 *
 * @list: heci io list pointer
 * @file: pointer to file object
 *
 * returns   returned a list entry on success, NULL on failure.
 */
struct heci_cb_private *heci_find_list_file_entry(
		struct heci_io_list *list,
		struct file *file)
{
	struct heci_cb_private *cb_pos = NULL;
	struct heci_cb_private *cb_next = NULL;

	if (list_empty(&list->cb_list))
		return NULL;

	list_for_each_entry_safe(cb_pos, cb_next, &list->cb_list, cb_list) {
		if (cb_pos->file_object == file)
			return cb_pos;
	}
	return NULL;
}

/**
 * heci_find_list_conn_entry - finds a list entry for current connection
 *
 * @list: heci io list pointer
 * @conn: connection data of file object
 *
 * returns   returned a list entry on success, NULL on failure.
 */
struct heci_cb_private *heci_find_list_conn_entry(
		struct heci_io_list *list,
		struct heci_connection *conn)
{
	struct heci_cb_private *cb_pos = NULL;
	struct heci_cb_private *cb_next = NULL;

	if (list_empty(&list->cb_list))
		return NULL;

	list_for_each_entry_safe(cb_pos, cb_next, &list->cb_list, cb_list) {
		if (cb_pos->conn == conn)
			return cb_pos;
	}
	return NULL;
}

/**
 * heci_find_read_entry - finds read cb entry for file
 *
 * @dev: Device object for our driver
 * @file: pointer to file object
 * @rcomplete: [out] HECI_COMPLETED if read operation is complete,
 *                   HECI_IN_PROGRESS if read in progress
 *                   HECI_WAIT_START if read not started
 *
 * returns   returned a list entry on success, NULL on failure.
 *
 * NOTE: result have to be freed with heci_put_io_cb
 */
struct heci_cb_private *heci_find_read_entry(struct iamt_heci_device *dev,
				struct file *file, __u8 *rcomplete)
{
	struct heci_cb_private *priv_cb = NULL;

	*rcomplete = HECI_WAIT_START;

	spin_lock_bh(&dev->rdcpl_lock);
	priv_cb = heci_get_io_cb(heci_find_list_file_entry(
			&dev->read_cmpl_list, file));
	spin_unlock_bh(&dev->rdcpl_lock);
	if (priv_cb != NULL) {
		*rcomplete = HECI_COMPLETED;
		return priv_cb;
	}

	spin_lock_bh(&dev->crrdl_lock);
	priv_cb = heci_get_io_cb(heci_find_list_file_entry(
			&dev->read_list, file));
	if (priv_cb != NULL)
		*rcomplete = HECI_IN_PROGRESS;
	else
		priv_cb = heci_get_io_cb(heci_find_list_file_entry(
			&dev->ctrl_wr_list, file));
	spin_unlock_bh(&dev->crrdl_lock);
	if ((priv_cb != NULL) &&
	    (priv_cb->major_file_operations == HECI_READ))
		return priv_cb;

	return NULL;
}

/**
 * heci_read_raw - read data to user
 *
 * @dev: Device object for our driver
 * @file_ext: private data of the file object
 * @priv_cb: heci call back struct
 * @ubuf: pointer to user data in user space
 * @length: data length to read
 * @offset: data read offset
 * @timeout: heci response expiration timeout
 * @needfree: [out] return 1 if callback need to be freed
 *
 * returns
 *  returned data length on success,
 *  zero if no data to read,
 *  negative on failure.
 */
int heci_read_raw(struct iamt_heci_device *dev,
		struct heci_file_priv *file_ext,
		struct heci_cb_private *priv_cb,
		char *ubuf, size_t length, loff_t *offset,
		unsigned long timeout, int *needfree)
{
	int rets = file_ext->status;

	*needfree = 0;

	if (priv_cb == NULL) {
		if (rets == 0)
			rets = -ENODEV;
		return rets;
	}

	if ((timeout > 0) && (priv_cb->read_time > 0) &&
	    ((get_seconds() - priv_cb->read_time) > timeout)) {
		*offset = 0;
		rets = -ETIMEDOUT;
		*needfree = 1;
		goto out;
	}

	if (*offset >= priv_cb->information) {
		if (atomic_read(&file_ext->r_state) == HECI_READ_COMPLETE) {
			*needfree = 1;
			return 0;
		}
		rets = -EMSGSIZE;
		*needfree = 1;
		goto out;
	}

	DBG("priv_cb->response_buffer size - %d\n",
		priv_cb->response_buffer.size);
	DBG("priv_cb->information - %lu\n", priv_cb->information);

	/* length is being turncated to PAGE_SIZE, however,
	   information size may be longer */
	length = (length < (priv_cb->information - *offset)) ?
			length : (priv_cb->information - *offset);

	rets = copy_to_user(ubuf,
			 priv_cb->response_buffer.data + *offset,
			 length);
	if (rets != 0) {
		rets = -EFAULT;
		goto out;
	}

	rets = length;
	*offset += length;
	if ((unsigned long)*offset >= priv_cb->information)
		*needfree = 1;

out:	DBG("end heci read rets= %d\n", rets);
	return rets;
}

/**
 * heci_have_slots - check if slots can contain heci message data.
 *
 * @slots: free slots.
 * @datalen: heci message data size in bytes.
 *
 * returns !=0 if slots can contain data, 0 - if not
 */
int heci_have_slots(__s32 *slots, __u32 datalen)
{
	return ((*slots * sizeof(__u32)) >=
		    (sizeof(struct heci_msg_hdr) + datalen));
}


/**
 * heci_start_read_flow - starts read by sending flow control
 *
 * @dev: Device object for our driver
 * @conn: connection data of the file object
 * @priv_cb: heci call back struct
 * @slots: free slots
 *
 * returns 0 on success, <0 on failure.
 */
int heci_start_read_flow(struct iamt_heci_device *dev,
			 struct heci_connection *conn,
			 struct heci_cb_private *priv_cb,
			 __s32 *slots)
{
	priv_cb->information = 0;

	if (slots == NULL) {
		if (atomic_cmpxchg(&dev->host_buffer_is_empty, 1, 0) != 1)
			return -EMESSAGE_NOT_COMPLETE;
	} else {
		if (!heci_have_slots(slots, sizeof(struct hbm_flow_control)))
			return -EMESSAGE_NOT_COMPLETE;
		*slots -= heci_len_to_slots(sizeof(struct hbm_flow_control));
	}

	if (!heci_send_flow_control(dev, conn)) {
		DBG("flow control failed\n");
		conn->status = -ENODEV;
		return -ENODEV;
	}

	conn->status = 0;
	DBG("flow control success\n");
	host_buffer_is_empty(dev, slots);

	if (conn->canceled == 1)
		return -ECANCELED;

	conn->reading_state = HECI_READING;
	conn->read_cb = priv_cb;
	conn->stall_timer = HECI_STALL_TIMER;
	DBG("STALL %d START_FLOW\n", HECI_STALL_TIMER);
	list_move_tail(&priv_cb->cb_list, &dev->read_list.cb_list);
	return 0;
}

/**
 * heci_len_to_slots - converts heci message data size to slots.
 *
 * @datalen: heci message data size in bytes.
 *
 * returns number of slots required to contain heci message data
 */
__u32 heci_len_to_slots(__u32 datalen)
{
	return (sizeof(struct heci_msg_hdr) + datalen + 3) / 4;
}

/**
 * _heci_write_buf_cmpl - write piece of data to heci.
 *
 * @dev: Device object for our driver.
 * @conn: connection data of the file object.
 * @buf: buffer with data to write.
 * @bufoff: write data offset in buffer.
 * @len: data length.
 * @complete: is last piece of data
 * @slots: free slots.
 *
 * returns  0 - OK; !=0 - error.
 */
static int _heci_write_buf_cmpl(struct iamt_heci_device *dev,
			struct heci_connection *conn,
			char *buf, unsigned long *bufoff, __u32 len,
			__u8 complete, __s32 *slots)
{
	struct heci_msg_hdr heci_hdr;
	int write_res;

	heci_hdr.host_addr = conn->host_client_id;
	heci_hdr.me_addr = conn->me_client_id;
	heci_hdr.length = len;
	heci_hdr.msg_complete = ((complete) ? 1 : 0);
	heci_hdr.reserved = 0;
	DBG("heci_hdr->msg_complete =%d\n", heci_hdr.msg_complete);
	DBG("bufoff =%lu\n", *bufoff);
	DBG("heci_hdr->length =%d\n", heci_hdr.length);
	if (slots != NULL)
		*slots -= heci_len_to_slots(heci_hdr.length);
	write_res = heci_write_message(dev, &heci_hdr,
			(unsigned char *) (buf + *bufoff), heci_hdr.length);
	if (write_res == 0) {
		conn->status = -ENODEV;
		return -ENODEV;
	}
	*bufoff += heci_hdr.length;
	DBG("bufoff =%lu\n", *bufoff);
	return 0;
}

/**
 * heci_write_buf - write data to heci.
 *
 * @dev: Device object for our driver.
 * @priv_cb: heci call back struct
 * @conn: connection data of the file object.
 * @wrspace: available write space.
 * @maxwrspace: maximum write space.
 * @slots: free slots.
 *
 * returns  0 - write completed; !=0 - not completed or error.
 */
int heci_write_buf(struct iamt_heci_device *dev,
			struct heci_cb_private *priv_cb,
			struct heci_connection *conn,
			__u32 wrspace, __u32 maxwrspace, __s32 *slots)
{
	int res;
	char *buf = priv_cb->request_buffer.data;
	unsigned long *bufoff = &(priv_cb->information);
	__u32 len = priv_cb->request_buffer.size - *bufoff;

	if (wrspace < (sizeof(struct heci_msg_hdr) + len)) {
		if ((maxwrspace != 0) && (wrspace != maxwrspace))
			return -EMESSAGE_WILL_SPLIT;

		res = _heci_write_buf_cmpl(dev, conn, buf, bufoff,
			(wrspace - sizeof(struct heci_msg_hdr)), 0, slots);
		if (res != 0)
			return res;
		return -EMESSAGE_NOT_COMPLETE;
	}

	res = _heci_write_buf_cmpl(dev, conn, buf, bufoff, len, 1, slots);
	if (res != 0)
		return res;
	flow_ctrl_reduce(dev, conn);
	conn->status = 0;
	return 0;
}

/**
 * heci_wd - process watchdog related operation.
 *
 * @dev: Device object for our driver.
 * @slots: free slots.
 *
 * returns  0, OK; otherwise, error.
 */
int heci_wd(struct iamt_heci_device *dev, __s32 *slots)
{
	__u32 len;
	struct heci_connection *conn = &dev->wd_conn;

	dev->wd_pending = 1;

	if (flow_ctrl_creds(dev, conn) == 0)
		return -EBUSY;

	conn->status = 0;
	if (dev->wd_timeout != 0)
		len = HECI_START_WD_DATA_SIZE;
	else
		len = HECI_WD_PARAMS_SIZE;
	if (slots == NULL) {
		if (atomic_cmpxchg(&dev->host_buffer_is_empty, 1, 0) != 1)
			return -EMESSAGE_NOT_COMPLETE;
	} else {
		if (!heci_have_slots(slots, len))
			return -EMESSAGE_NOT_COMPLETE;
		*slots -= heci_len_to_slots(len);
	}

	if (heci_send_wd(dev)) {
		flow_ctrl_reduce(dev, conn);
		if (dev->wd_stop == 1)
			dev->wd_stop = 2;
	} else
		conn->status = -ENODEV;

	dev->wd_pending = 0;
	if (dev->wd_timeout != 0)
		dev->wd_due_counter = 2;
	else
		dev->wd_due_counter = 0;

	return 0;
}

